/********************************************************************

noiseclock: sumarize time of mp3 and other sound formats
Copyright (C) 2004 Jack Chatterton <chattj AT postreal.org>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  
02111-1307, USA.

*********************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <dirent.h>
#include <unistd.h>
#include <taglib/tag_c.h>
/* Stack of directory entries, a customized data structure */
#include "dirstack.h"
/* Pretty-print output routines */
#include "time_formatter.h"
/* Utilities include safe memory allocation routines and the calls which
 turn off stderr. */
#include "utilities.h"
/* Home-made fast replacement for ftw. */
#include "dirwalk.h"

/* 999999 is equal to infinity, for exceedingly low values of infinity. */
#define INFINITY 999999
#define MAX_FILENAME_SIZE 2048
#define MAX_ALLOWED_DEPTH 64

char *error_text = "\
 [OPTION]... [FILE]... \n\
Noiseclock is designed to summarize the amount of time, in any combination\n\
of days, minutes, hours, and/or seconds, needed to play a given file or\n\
directory of files. It supports any of the sound formats supported by TagLib,\n\
which are currently MP3, OGG, and FLAC compressed files.\n\
\n\
  -a,     Output times for all files scanned, not just directories.\n\
  -c,     Print grand total of all counts.\n\
  -f STR, Format time output using a specified format string. STR takes \n\
          the form of any combination of the characters 'd', 'h', 'm', and \n\
          's' which stand for days, hours, minutes, and seconds \n\
          respectively. Output will be the corresponding numbers, in order \n\
          and with repeats. Whitespace terminates the string, and \n\
          meaningless characters will cause the program to error out.\n\
  -n,     Do not display time format hints, so output is not human readable.\n\
  -q,     Allow output of TagLib file errors. Suppressed by default.\n\
  -x N,   Print directories to a maximum depth of N.\n\
\n\
Report bugs to <chattj AT postreal.org>.\n";

/* Globals */
int high_depth = 0;
int low_depth = INFINITY;
char *program_name;
char *target_name = NULL;

/* Program argument globals. Like, how the program will run. These
 are updated in process_args() */
/* Directory display depth */
int max_depth = MAX_ALLOWED_DEPTH;
/* Allow human-readable formatting by default. */
int human_readable = 1;
/* Whether or not to suppress worthless TagLib errors. */
int allow_stderr = 0;
/* During directory crawl, whether or not to show each filename scanned */
int show_filenames = 0;
/* Grand total to be output if the user choses that option */
int print_grand_total = 0;
int grand_total = 0;

/*
 Callback to be called for each file by ftw. This is responsible for
 building the list of directories that will be searched.
 name is the full pathname
 status is a stat structure, "man 2 stat" for knowledge. We will not
 be using it here.
 type is the type of file, set up in ftw.h. It can be:
 	FTW_F - normal file
	FTW_D - directory
	FTW_DNR - unreadable directory
	FTW_SL - symbolic link
	FTW_NS - failure (unreadable?) on item which is not symbolic link
*/
/*
int maketree_callback(const char *name, const struct stat *status, int type) {
	struct dir_entry *insertme;
	insertme = (struct dir_entry *) safe_malloc(sizeof(struct dir_entry));
	if(type == FTW_D) {
		insertme->dirname = (char *) safe_malloc(sizeof(char[strlen(name)+1]));
		strcpy(insertme->dirname, name);
		insertme->total = 0;
		insertme->depth = countslashes((char *) name);
		if (insertme->depth < low_depth)
			low_depth = insertme->depth;
		if (insertme->depth > high_depth)
			high_depth = insertme->depth;
		dirpush(insertme);
	} else {
        safe_free(insertme);
    }
    */
	/* Always return 0 so ftw will run forever */
//	return 0;
//}

/* New callback to use with dirwalk.h. HAZARD: This is fragile in that it 
assumes it will be fed only directories, and inserts them accordingly. */
int maketree_callback(const char *name) {
    struct dir_entry *insertme;
    insertme = (struct dir_entry *) safe_malloc(sizeof(struct dir_entry));
    insertme->dirname = (char *) safe_malloc(sizeof(char[strlen(name)+1]));
    strcpy(insertme->dirname, name);
    insertme->total = 0;
    insertme->depth = countslashes((char *) name);
    if (insertme->depth < low_depth)
        low_depth = insertme->depth;
    if (insertme->depth > high_depth)
        high_depth = insertme->depth;
    dirpush(insertme);
    return 0;
}

/* Return -1 if we need to stop the program. */
int process_args(int argc, char *argv[]) {
	int c, returnme;
	char *formatstr = NULL;
	/* Set program's name */
	program_name = argv[0];
	/* Cheat key:
	 d = day, h = hour, m = minute, s = second. */
	while((c = getopt(argc, argv, "acqnf:x:")) != -1) {
		switch(c) {
			case '?':
				printf("Usage: %s%s",program_name,error_text);
				return -1;
			case 'f':
				formatstr = (char *) safe_malloc(sizeof(optarg));
				strcpy(formatstr, optarg);
				break;
			case 'x':
				max_depth = atoi(optarg);
				break;
			case 'n':
				human_readable = 0;
				break;
			case 'q':
				allow_stderr = 1;
				break;
			case 'a':
				show_filenames = 1;
				break;
            case 'c':
                print_grand_total = 1;
                break;
		}
	}
	/* default to hours, minutes, seconds format */
	if(formatstr) {
		returnme = setup_formatter(formatstr);
        safe_free(formatstr);
        return returnme;
    } else
		return setup_formatter("hms");
}

int getseconds(char *filename) {
    int seconds = 0;
	const TagLib_AudioProperties *props;
	TagLib_File *file;
    file = taglib_file_new(filename);
    /* Only deal with the file if it isn't null */
    if(file) {
        props = taglib_file_audioproperties(file);
        seconds = taglib_audioproperties_length(props);
        /* If getseconds is processing a file, it obviously must be part of
         the grand total. Thus, this calculation can go here and here only.
         This means that files which are looked at redundantly will 
         contribute twice to the grand total, which behavior makes sense. */
        grand_total += seconds;
        if(show_filenames) 
            printf("%s%s\n", time_format(seconds, human_readable), filename);
        taglib_file_free(file);
    }
	file = NULL;
    return seconds;
}    

void mt_dirloop() {
	int depths[MAX_ALLOWED_DEPTH];
	int size = getsize();
	struct dir_entry *tempentry;
	char *filename;
	int seconds = 0, curdepth = 0;
	DIR *dp;
	struct dirent *ep;
	int i,j;
	/* PROBLEM: chars may be frickin huge on some systems, so this 
	 number may not be nearly big enough. Whatever. Here in real
	 life it almost certainly will be. */
	filename = (char *) safe_malloc(MAX_FILENAME_SIZE);
	
	/* Create array to hold temporary accumulations for each depth. */
	for(i = 0; i < (high_depth + 1); i++) 
		depths[i] = 0;

	/* Loop over all directories */
	for(i = 1; i <= size; i++) {
		tempentry = dirpop();
		/* Here we look at the directory, see dirent.h for info. */
		dp = opendir(tempentry->dirname);
		curdepth = tempentry->depth;
		if(dp != NULL) {
            /* The compiler suggests these extra parentheses (?) */
			while((ep = readdir(dp))) {
				/* Uhm, the type can be lots of things. One important
				 thing it can be is DT_REG, a regular file. Another is 
				 DT_LNK, presumably a softlink. Everything else is 
				 unimportant here. */
				if(ep->d_type == DT_REG) {
					/* Sweet jesus c string manipulation is ugly. Is there
					 a better way to do this? */
					strcpy(filename, tempentry->dirname);
					strcat(filename, "/");
					strcat(filename, ep->d_name);
                    /* Get actual number of seconds for file */
                    seconds = getseconds(filename);
					/* Set summations for all appropriate depths. */
					for(j = 0; j <= curdepth; j++)
						depths[j] += seconds;
					seconds = 0;
				}
			}
		} else {
			printf("Couldn't open directory %s.\n",tempentry->dirname);
		}
		closedir(dp);
        /* This clause makes the max-depth argument possible */
        if((curdepth - low_depth) <= max_depth)
    		printf("%s%s\n", time_format(depths[curdepth], human_readable), tempentry->dirname);
		depths[curdepth] = 0;
		safe_free(tempentry->dirname);
		safe_free(tempentry);
	}
    safe_free(filename);
}

void process_target(char *target) {
	DIR *test;
    /* Check to see whether user wants to look at a file or a directory */
    test = opendir(target);
    if(test) {    
        safe_free(test);
    	
        /* Build stack of all directories. This is a kind of ghetto 
	     way to do a depth-first search. target is defined in
    	 process_args. */
//	    ftw(target,maketree_callback,1);
        dirwalk(target, maketree_callback);

	    /* Do the main loop, which prints all info. */
    	mt_dirloop();
	} else {
        /* Deal with a single file and exit. */
        show_filenames = 0;
		printf("%s%s\n", time_format(getseconds(target), human_readable), target);
    } 
}


int main(int argc, char *argv[]) {
	/* Process command line, and exit if it's broken */
	if(process_args(argc, argv) == -1)
		return -1;
	
	/* Disallow stderr if necessary */
	if(!(allow_stderr))
		suppress_errors();
	
	/* Initialize stack of directories. See dirstack.h for more info. */
	create_dirstack();

	/* Set directory which will be searched. optind is set in 
	 process_args to be number of actual input (i.e. anything that's
	 not a directive) */
	target_name = (char *) safe_malloc(MAX_FILENAME_SIZE);
	if(optind < argc) 
		while(optind < argc) {
			strcpy(target_name, argv[optind]);
            process_target(target_name);
			optind++;
		}
	else {
		/* Default is to search under current directory */
		strcpy(target_name, ".");
		process_target(target_name);
	}

    /* Print grand total if user wants that to happen. */
    if(print_grand_total)
        printf("%s%s\n", time_format(grand_total, human_readable), "total");
	
	/* Success! */
	return 0;
}
